/*
 *	pycpu.c -- CPU information query library for Python
 *	Based on cpuinfo.c utility from MPlayer
 *
 * pycpu is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * pycpu is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with MPlayer; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/
#include <stdio.h>
#include <sys/time.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <malloc.h>

#ifdef BUILD_LIB
	/* DLL/.so export */
	#define EXPORT __declspec(dllexport)
#else
	/* EXE import */
	#define EXPORT __declspec(dllimport)
#endif

#ifdef __MINGW32__
	#define MISSING_USLEEP
	#include <windows.h>
	#define sleep(t) Sleep(1000*t);
#endif

#ifdef M_UNIX
	typedef long long int64_t;
	#define MISSING_USLEEP
#else
	#include <inttypes.h>
#endif

#define CPUID_FEATURE_DEF(bit, desc) \
    { bit, desc }
	
typedef struct cpuid_regs {
    unsigned int eax;
    unsigned int ebx;
    unsigned int ecx;
    unsigned int edx;
} cpuid_regs_t;

EXPORT typedef struct t_cpu_cap {
	int id;
	int bit;
	char *desc;
	struct t_cpu_cap *next;
} cpu_cap;

typedef struct t_cpu_info {
	char vendor_id[13];
	unsigned int max_cpuid;
	unsigned int max_ext_cpuid;
	unsigned int amd_flags;
	unsigned int amd_flags2;
	const char *model_name;
	char processor_name[49];
	unsigned int family;
	unsigned int model;
	unsigned int stepping;
	double clock_mhz;
	double clock_ghz;
} cpu_info;

EXPORT cpu_info ProcInfo;
EXPORT int numCaps;
EXPORT cpu_cap* ProcCaps;
int libStarted;

EXPORT __stdcall cpu_cap getCap(int capNum) {	
	cpu_cap *current;
	current = ProcCaps;
	
	if (current->id == capNum) return *current;
	
	while (current != NULL) {
		if (current->id == capNum) {
			return *current;
		} else {
			current = current->next;
		}
	}

	return *current;
}

static void addCap(int bit, char *desc) {
	cpu_cap *thiscap;
	cpu_cap *current;
	
	if ( (thiscap = (cpu_cap *)malloc(sizeof(cpu_cap))) == NULL) return;
	memset(thiscap, 0, sizeof(cpu_cap));
	thiscap->id = numCaps;
	thiscap->bit = bit;
	thiscap->desc = desc;
	thiscap->next = NULL;
	
	/*
		Join the planet to the list
	*/
	if (ProcCaps == NULL) {
		ProcCaps = thiscap;
	} else {
		current = ProcCaps;
		while (current->next != NULL) {
			current = current->next;
		}
		current->next = thiscap;
	}
	
	numCaps++;
}

static cpuid_regs_t cpuid(int func) {
    cpuid_regs_t regs;
#define CPUID   ".byte 0x0f, 0xa2; "
#ifdef __x86_64__
    __asm__("mov %%rbx, %%rsi\n\t"
#else
    __asm__("mov %%ebx, %%esi\n\t"
#endif
            CPUID"\n\t"
#ifdef __x86_64__
            "xchg %%rsi, %%rbx\n\t"
#else
            "xchg %%esi, %%ebx\n\t"
#endif
            : "=a" (regs.eax), "=S" (regs.ebx), "=c" (regs.ecx), "=d" (regs.edx)
            : "0" (func));
    return regs;
}

static int64_t rdtsc(void) {
    uint32_t hi, lo;
#define RDTSC   ".byte 0x0f, 0x31; "
    __asm__ volatile (RDTSC : "=a"(lo), "=d"(hi) : );
    return (uint64_t) hi << 32 | lo;
}

static const char* brandname(int i) {
    static const char* brandmap[] = {
        NULL,
        "Intel(R) Celeron(R) processor",
        "Intel(R) Pentium(R) III processor",
        "Intel(R) Pentium(R) III Xeon(tm) processor",
        "Intel(R) Pentium(R) III processor",
        NULL,
        "Mobile Intel(R) Pentium(R) III processor-M",
        "Mobile Intel(R) Celeron(R) processor"
    };

    if (i >= sizeof(brandmap))
        return NULL;
    else
        return brandmap[i];
}

static void store32(char *d, unsigned int v) {
    d[0] =  v        & 0xff;
    d[1] = (v >>  8) & 0xff;
    d[2] = (v >> 16) & 0xff;
    d[3] = (v >> 24) & 0xff;
}

void lib_start(void) {
	numCaps = 0;
	ProcInfo.model_name = NULL;
	ProcInfo.clock_mhz = 0.00;
	ProcInfo.clock_ghz = 0.00;
	ProcInfo.amd_flags = 0;
	ProcInfo.amd_flags2 = 0;
	ProcCaps = NULL;
	libStarted = 1;
}

void destroy_caps(void) {
	cpu_cap *current;
	cpu_cap *temp;
    current = ProcCaps;
	
    while (current != NULL) {
        temp = current->next;
        free(current);
        current = temp;
    }
	
    ProcCaps = NULL;
	numCaps = 0;
}

EXPORT __stdcall void lib_shutdown(void) {
	destroy_caps();
}

EXPORT __stdcall void getCPUInfo(void) {
	cpuid_regs_t regs, regs_ext;
    int i;
	
	if (!libStarted) lib_start();
	
    regs = cpuid(0);
    ProcInfo.max_cpuid = regs.eax;
    /*printf("%d CPUID function codes\n", ProcInfo.max_cpuid+1);*/

    store32(ProcInfo.vendor_id+0, regs.ebx);
    store32(ProcInfo.vendor_id+4, regs.edx);
    store32(ProcInfo.vendor_id+8, regs.ecx);
    ProcInfo.vendor_id[12] = 0;
	
    regs_ext = cpuid((1<<31) + 0);
    ProcInfo.max_ext_cpuid = regs_ext.eax;
    if (ProcInfo.max_ext_cpuid >= (1<<31) + 1) {
        regs_ext = cpuid((1<<31) + 1);
        ProcInfo.amd_flags = regs_ext.edx;
        ProcInfo.amd_flags2 = regs_ext.ecx;

        if (ProcInfo.max_ext_cpuid >= (1<<31) + 4) {
            for (i = 2; i <= 4; i++) {
                regs_ext = cpuid((1<<31) + i);
                store32(ProcInfo.processor_name + (i-2)*16, regs_ext.eax);
                store32(ProcInfo.processor_name + (i-2)*16 + 4, regs_ext.ebx);
                store32(ProcInfo.processor_name + (i-2)*16 + 8, regs_ext.ecx);
                store32(ProcInfo.processor_name + (i-2)*16 + 12, regs_ext.edx);
            }
            ProcInfo.processor_name[48] = 0;
            
			ProcInfo.model_name = ProcInfo.processor_name;
			
            while (*ProcInfo.model_name == ' ')
                ProcInfo.model_name++;
        }
    } 

    if (ProcInfo.max_cpuid >= 1) {
        static struct {
            int bit;
            char *desc;
        } cap[] = {
            CPUID_FEATURE_DEF(0, "fpu"),
            CPUID_FEATURE_DEF(1, "vme"),
            CPUID_FEATURE_DEF(2, "de"),
            CPUID_FEATURE_DEF(3, "pse"),
            CPUID_FEATURE_DEF(4, "tsc"),
            CPUID_FEATURE_DEF(5, "msr"),
            CPUID_FEATURE_DEF(6, "pae"),
            CPUID_FEATURE_DEF(7, "mce"),
            CPUID_FEATURE_DEF(8, "cx8"),
            CPUID_FEATURE_DEF(9, "apic"),
            CPUID_FEATURE_DEF(11, "sep"),
            CPUID_FEATURE_DEF(12, "mtrr"),
            CPUID_FEATURE_DEF(13, "pge"),
            CPUID_FEATURE_DEF(14, "mca"),
            CPUID_FEATURE_DEF(15, "cmov"),
            CPUID_FEATURE_DEF(16, "pat"),
            CPUID_FEATURE_DEF(17, "pse36"),
            CPUID_FEATURE_DEF(18, "pn"),
            CPUID_FEATURE_DEF(19, "clflush"),
            CPUID_FEATURE_DEF(21, "dts"),
            CPUID_FEATURE_DEF(22, "acpi"),
            CPUID_FEATURE_DEF(23, "mmx"),
            CPUID_FEATURE_DEF(24, "fxsr"),
            CPUID_FEATURE_DEF(25, "sse"),
            CPUID_FEATURE_DEF(26, "sse2"),
            CPUID_FEATURE_DEF(27, "ss"),
            CPUID_FEATURE_DEF(28, "ht"),
            CPUID_FEATURE_DEF(29, "tm"),
            CPUID_FEATURE_DEF(30, "ia64"),
            CPUID_FEATURE_DEF(31, "pbe"),
            { -1 }
        };
        static struct {
            int bit;
            char *desc;
        } cap2[] = {
            CPUID_FEATURE_DEF(0, "pni"),
            CPUID_FEATURE_DEF(1, "pclmulqdq"),
            CPUID_FEATURE_DEF(2, "dtes64"),
            CPUID_FEATURE_DEF(3, "monitor"),
            CPUID_FEATURE_DEF(4, "ds_cpl"),
            CPUID_FEATURE_DEF(5, "vmx"),
            CPUID_FEATURE_DEF(6, "smx"),
            CPUID_FEATURE_DEF(7, "est"),
            CPUID_FEATURE_DEF(8, "tm2"),
            CPUID_FEATURE_DEF(9, "ssse3"),
            CPUID_FEATURE_DEF(10, "cid"),
            CPUID_FEATURE_DEF(12, "fma"),
            CPUID_FEATURE_DEF(13, "cx16"),
            CPUID_FEATURE_DEF(14, "xtpr"),
            CPUID_FEATURE_DEF(15, "pdcm"),
            CPUID_FEATURE_DEF(18, "dca"),
            CPUID_FEATURE_DEF(19, "sse4_1"),
            CPUID_FEATURE_DEF(20, "sse4_2"),
            CPUID_FEATURE_DEF(21, "x2apic"),
            CPUID_FEATURE_DEF(22, "movbe"),
            CPUID_FEATURE_DEF(23, "popcnt"),
            CPUID_FEATURE_DEF(25, "aes"),
            CPUID_FEATURE_DEF(26, "xsave"),
            CPUID_FEATURE_DEF(27, "osxsave"),
            CPUID_FEATURE_DEF(28, "avx"),
            { -1 }
        };
        static struct {
            int bit;
            char *desc;
        } cap_amd[] = {
            CPUID_FEATURE_DEF(11, "syscall"),
            CPUID_FEATURE_DEF(19, "mp"),
            CPUID_FEATURE_DEF(20, "nx"),
            CPUID_FEATURE_DEF(22, "mmxext"),
            CPUID_FEATURE_DEF(25, "fxsr_opt"),
            CPUID_FEATURE_DEF(26, "pdpe1gb"),
            CPUID_FEATURE_DEF(27, "rdtscp"),
            CPUID_FEATURE_DEF(29, "lm"),
            CPUID_FEATURE_DEF(30, "3dnowext"),
            CPUID_FEATURE_DEF(31, "3dnow"),
            { -1 }
        };
		
        static struct {
            int bit;
            char *desc;
        } cap_amd2[] = {
            CPUID_FEATURE_DEF(0, "lahf_lm"),
            CPUID_FEATURE_DEF(1, "cmp_legacy"),
            CPUID_FEATURE_DEF(2, "svm"),
            CPUID_FEATURE_DEF(3, "extapic"),
            CPUID_FEATURE_DEF(4, "cr8_legacy"),
            CPUID_FEATURE_DEF(5, "abm"),
            CPUID_FEATURE_DEF(6, "sse4a"),
            CPUID_FEATURE_DEF(7, "misalignsse"),
            CPUID_FEATURE_DEF(8, "3dnowprefetch"),
            CPUID_FEATURE_DEF(9, "osvw"),
            CPUID_FEATURE_DEF(10, "ibs"),
            CPUID_FEATURE_DEF(11, "sse5"),
            CPUID_FEATURE_DEF(12, "skinit"),
            CPUID_FEATURE_DEF(13, "wdt"),
            { -1 }
        };
        
        regs = cpuid(1);
        ProcInfo.family = (regs.eax >> 8) & 0xf;
        ProcInfo.model = (regs.eax >> 4) & 0xf;
        ProcInfo.stepping = regs.eax & 0xf;

        if (ProcInfo.family == 0xf)
            ProcInfo.family += (regs.eax >> 20) & 0xff;
        if (ProcInfo.family == 0xf || ProcInfo.family == 6)
            ProcInfo.model += ((regs.eax >> 16) & 0xf) << 4;


        if (strstr(ProcInfo.vendor_id, "Intel") && !ProcInfo.model_name) {
            if (ProcInfo.family == 6 && ProcInfo.model == 0xb && ProcInfo.stepping == 1)
                ProcInfo.model_name = "Intel (R) Celeron (R) processor";
            else
                ProcInfo.model_name = brandname(regs.ebx & 0xf);
        }
		
        for (i = 0; cap[i].bit >= 0; i++) {
            if (regs.edx & (1 << cap[i].bit))
				addCap(cap[i].bit, cap[i].desc);
        }
        for (i = 0; cap2[i].bit >= 0; i++) {
            if (regs.ecx & (1 << cap2[i].bit))
				addCap(cap2[i].bit, cap2[i].desc);
        }
		
        /* k6_mtrr is supported by some AMD K6-2/K6-III CPUs but
           it is not indicated by a CPUID feature bit, so we
           have to check the family, model and stepping instead. */
        if (strstr(ProcInfo.vendor_id, "AMD") &&
            ProcInfo.family == 5 &&
            (ProcInfo.model >= 9 || (ProcInfo.model == 8 && ProcInfo.stepping >= 8)))
				addCap(0, "k6_mtrr");
        /* similar for cyrix_arr. */
        if (strstr(ProcInfo.vendor_id, "Cyrix") &&
            (ProcInfo.family == 5 && (ProcInfo.model < 4 || ProcInfo.family == 6)))
			addCap(0, "cyrix_arr");
		
        /* as well as centaur_mcr. */
        if (strstr(ProcInfo.vendor_id, "Centaur") &&
            ProcInfo.family == 5)
			addCap(0, "centaur_mcr");

        for (i = 0; cap_amd[i].bit >= 0; i++) {
            if (ProcInfo.amd_flags & (1 << cap_amd[i].bit)) {
				addCap(cap_amd[i].bit, cap_amd[i].desc);
            }
        }
        for (i = 0; cap_amd2[i].bit >= 0; i++) {
            if (ProcInfo.amd_flags2 & (1 << cap_amd2[i].bit)) {
				addCap(cap_amd2[i].bit, cap_amd2[i].desc);
            }
        }
        
        if (regs.edx & (1 << 4)) {
            int64_t tsc_start, tsc_end;
            struct timeval tv_start, tv_end;
            int usec_delay;

            tsc_start = rdtsc();
            gettimeofday(&tv_start, NULL);
#ifdef  MISSING_USLEEP
            sleep(1);
#else
            usleep(100000);
#endif
            tsc_end = rdtsc();
            gettimeofday(&tv_end, NULL);

            usec_delay = 1000000 * (tv_end.tv_sec - tv_start.tv_sec)
                + (tv_end.tv_usec - tv_start.tv_usec);

			ProcInfo.clock_mhz = (double)(tsc_end-tsc_start) / usec_delay;
		   
		    /* Get GHz speed of clock if the MHz speed is
			   >= 1000MHz. */
		    if (ProcInfo.clock_mhz >= 1000.00)
				ProcInfo.clock_ghz = (double)(ProcInfo.clock_mhz / 1000.00);
        }
    }
}